@@ -1,4 +1,3 @@
-
 # -*- coding: utf-8 -*-
 # Copyright (c) 2011-2021, Albert Zeyer, www.az2000.de
 # All rights reserved.
@@ -41,14 +40,16 @@ See these functions:
 - get_current_frame
 - dump_all_thread_tracebacks
 - install
+- setup_all
 - replace_traceback_format_tb
+- replace_traceback_print_tb
 
 Although there might be a few more useful functions, thus we export all of them.
 
 Also see the demo/tests at the end.
 """
 
-from __future__ import print_function
+from __future__ import annotations
 
 import sys
 import os
@@ -57,6 +58,7 @@ import threading
 import keyword
 import inspect
 import contextlib
+
 try:
     import typing
 except ImportError:
@@ -65,8 +67,10 @@ except ImportError:
 try:
     from traceback import StackSummary, FrameSummary
 except ImportError:
+
     class _Dummy:
         pass
+
     StackSummary = FrameSummary = _Dummy
 
 # noinspection PySetFunctionToLiteral,SpellCheckingInspection
@@ -79,7 +83,7 @@ try:
     # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
     unicode
 except NameError:  # Python3
-    unicode = str   # Python 3 compatibility
+    unicode = str  # Python 3 compatibility
 
 try:
     # noinspection PyUnresolvedReferences,PyUnboundLocalVariable
@@ -138,7 +142,7 @@ def parse_py_statement(line):
         elif state == 1:  # string via "
             if c == "\\":
                 state = 4
-            elif c == "\"":
+            elif c == '"':
                 yield "str", cur_token
                 cur_token = ""
                 state = 0
@@ -205,8 +209,8 @@ def grep_full_py_identifiers(tokens):
         i += 1
         if token_type != "id":
             continue
-        while i+1 < len(tokens) and tokens[i] == ("op", ".") and tokens[i+1][0] == "id":
-            token += "." + tokens[i+1][1]
+        while i + 1 < len(tokens) and tokens[i] == ("op", ".") and tokens[i + 1][0] == "id":
+            token += "." + tokens[i + 1][1]
             i += 2
         if token == "":
             continue
@@ -228,7 +232,8 @@ def set_linecache(filename, source):
     :return: nothing
     """
     import linecache
-    linecache.cache[filename] = None, None, [line+'\n' for line in source.splitlines()], filename
+
+    linecache.cache[filename] = None, None, [line + "\n" for line in source.splitlines()], filename
 
 
 # noinspection PyShadowingBuiltins
@@ -293,6 +298,7 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
     try:
         # noinspection PyPackageRequirements
         import IPython
+
         have_ipython = True
     except ImportError:
         have_ipython = False
@@ -302,10 +308,13 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
         try:
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             from IPython.core.debugger import Pdb
+
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             from IPython.terminal.debugger import TerminalPdb
+
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             from IPython.terminal.ipapp import TerminalIPythonApp
+
             ipapp = TerminalIPythonApp.instance()
             ipapp.interact = False  # Avoid output (banner, prints)
             ipapp.initialize(argv=[])
@@ -317,7 +326,7 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
                 """
                 Run the IPython shell.
                 """
-                pdb_obj.interaction(None, traceback=traceback)
+                pdb_obj.interaction(None, traceback)
 
         except Exception:
             print("IPython Pdb exception:")
@@ -328,19 +337,20 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
         try:
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             import IPython
+
             # noinspection PyPackageRequirements,PyUnresolvedReferences
             import IPython.terminal.embed
 
-            class DummyMod(object):
+            class DummyMod:
                 """Dummy module"""
+
             module = DummyMod()
             module.__dict__ = user_global_ns
             module.__name__ = "_DummyMod"
             if "__name__" not in user_ns:
                 user_ns = user_ns.copy()
                 user_ns["__name__"] = "_DummyUserNsMod"
-            ipshell = IPython.terminal.embed.InteractiveShellEmbed.instance(
-                user_ns=user_ns, user_module=module)
+            ipshell = IPython.terminal.embed.InteractiveShellEmbed.instance(user_ns=user_ns, user_module=module)
         except Exception:
             print("IPython not available:")
             better_exchook(*sys.exc_info(), autodebugshell=False, file=sys.stdout)
@@ -354,6 +364,7 @@ def debug_shell(user_ns, user_global_ns, traceback=None, execWrapper=None):
         print("Use simple pdb debug shell:")
         if traceback:
             import pdb
+
             pdb.post_mortem(traceback)
         else:
             simple_debug_shell(user_global_ns, user_ns)
@@ -454,6 +465,7 @@ def get_source_code(filename, lineno, module_globals=None):
     :rtype: str
     """
     import linecache
+
     linecache.checkcache(filename)
     source_code = linecache.getline(filename, lineno, module_globals)
     # In case of a multi-line statement, lineno is usually the last line.
@@ -474,7 +486,7 @@ def get_source_code(filename, lineno, module_globals=None):
             end_line += 1
             if end_line > len(lines):  # 1-indexed
                 break
-        source_code = "".join(lines[start_line - 1:end_line])  # 1-indexed
+        source_code = "".join(lines[start_line - 1 : end_line])  # 1-indexed
     return source_code
 
 
@@ -485,6 +497,7 @@ def str_visible_len(s):
     :rtype: int
     """
     import re
+
     # via: https://github.com/chalk/ansi-regex/blob/master/index.js
     s = re.sub("[\x1b\x9b][\\[()#;?]*(?:[0-9]{1,4}(?:;[0-9]{0,4})*)?[0-9A-PRZcf-nqry=><]", "", s)
     return len(s)
@@ -510,7 +523,7 @@ def get_indent_prefix(s):
     :return: the indent spaces of s
     :rtype: str
     """
-    return s[:len(s) - len(s.lstrip())]
+    return s[: len(s) - len(s.lstrip())]
 
 
 def get_same_indent_prefix(lines):
@@ -540,7 +553,7 @@ def remove_indent_lines(s):
     prefix = get_same_indent_prefix(lines)
     if prefix is None:  # not in expected format. just lstrip all lines
         return "".join([line.lstrip() for line in lines])
-    return "".join([line[len(prefix):] for line in lines])
+    return "".join([line[len(prefix) :] for line in lines])
 
 
 def replace_tab_indent(s, replace="    "):
@@ -550,7 +563,7 @@ def replace_tab_indent(s, replace="    "):
     :rtype: str
     """
     prefix = get_indent_prefix(s)
-    return prefix.replace("\t", replace) + s[len(prefix):]
+    return prefix.replace("\t", replace) + s[len(prefix) :]
 
 
 def replace_tab_indents(s, replace="    "):
@@ -585,8 +598,9 @@ class Color:
     Helper functions provided to perform terminal coloring.
     """
 
-    ColorIdxTable = {k: i for (i, k) in enumerate([
-        "black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"])}
+    ColorIdxTable = {
+        k: i for (i, k) in enumerate(["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"])
+    }
 
     @classmethod
     def get_global_color_enabled(cls):
@@ -718,7 +732,7 @@ class Color:
             elif state == 1:  # string via "
                 if c == "\\":
                     state = 4
-                elif c == "\"":
+                elif c == '"':
                     color_args[i] = {}
                     state = 0
             elif state == 2:  # string via '
@@ -768,6 +782,7 @@ class DomTerm:
         :rtype: bool
         """
         import os
+
         if cls._is_domterm is not None:
             return cls._is_domterm
         if not os.environ.get("DOMTERM"):
@@ -800,7 +815,7 @@ class DomTerm:
         """
         :param io.TextIOBase|io.StringIO file:
         """
-        file.write("\033]114;\"│\"\007")
+        file.write('\033]114;"│"\007')
 
     # noinspection PyMethodMayBeStatic
     def hide_button(self, file=sys.stdout):
@@ -826,6 +841,7 @@ class DomTerm:
             and collect the data during the context (in the `with` block).
         """
         import io
+
         if hidden_stream is None:
             hidden_stream = sys.stdout
         assert isinstance(hidden_stream, io.IOBase)
@@ -881,6 +897,7 @@ class DomTerm:
         :rtype: str
         """
         import io
+
         output_buf = io.StringIO()
         self.fold_text(prefix=prefix, hidden=hidden, file=output_buf, **kwargs)
         return output_buf.getvalue()
@@ -903,7 +920,7 @@ def is_at_exit():
     return False
 
 
-class _Output:
+class _OutputLinesCollector:
     def __init__(self, color):
         """
         :param Color color:
@@ -944,6 +961,7 @@ class _Output:
         self.lines, new_lines = old_lines, self.lines  # recover self.lines
         hidden_text = "".join(new_lines)
         import io
+
         output_buf = io.StringIO()
         prefix = ""
         while line[:1] == " ":
@@ -959,6 +977,11 @@ class _Output:
         :param int depth_limit:
         :rtype: str
         """
+        # We want to exclude array types from Numpy, TensorFlow, PyTorch, etc.
+        # Getting __getitem__ or __len__ on them even could lead to unexpected results or deadlocks,
+        # depending on the context (e.g. inside a TF session run, extending the graph is unexpected).
+        if hasattr(obj, "shape"):
+            return ""
         s = []
         if hasattr(obj, "__len__"):
             # noinspection PyBroadException
@@ -995,7 +1018,7 @@ class _Output:
                 s = self.color.py_syntax_highlight(s)
                 s = self.dom_term.fold_text_string("", s)
             else:
-                s = s[:limit - 3]  # cut before syntax highlighting, to avoid missing color endings
+                s = s[: limit - 3]  # cut before syntax highlighting, to avoid missing color endings
                 s = self.color.py_syntax_highlight(s)
                 s += "..."
         else:
@@ -1021,7 +1044,7 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
     :rtype: list[str]
     """
     color = Color(enable=with_color)
-    output = _Output(color=color)
+    output = _OutputLinesCollector(color=color)
 
     def format_filename(s):
         """
@@ -1030,9 +1053,10 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
         """
         base = os.path.basename(s)
         return (
-            color('"' + s[:-len(base)], color.fg_colors[2]) +
-            color(base, color.fg_colors[2], bold=True) +
-            color('"', color.fg_colors[2]))
+            color('"' + s[: -len(base)], color.fg_colors[2])
+            + color(base, color.fg_colors[2], bold=True)
+            + color('"', color.fg_colors[2])
+        )
 
     format_py_obj = output.pretty_print
     if tb is None:
@@ -1054,9 +1078,9 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
     isframe = inspect.isframe
     if withTitle:
         if isframe(tb) or is_stack_summary(tb):
-            output(color('Traceback (most recent call first):', color.fg_colors[0]))
+            output(color("Traceback (most recent call first):", color.fg_colors[0]))
         else:  # expect traceback-object (or compatible)
-            output(color('Traceback (most recent call last):', color.fg_colors[0]))
+            output(color("Traceback (most recent call last):", color.fg_colors[0]))
     if with_vars is None and is_at_exit():
         # Better to not show __repr__ of some vars, as this might lead to crashes
         # when native extensions are involved.
@@ -1079,7 +1103,7 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
     # noinspection PyBroadException
     try:
         if limit is None:
-            if hasattr(sys, 'tracebacklimit'):
+            if hasattr(sys, "tracebacklimit"):
                 limit = sys.tracebacklimit
         n = 0
         _tb = tb
@@ -1141,20 +1165,25 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                 lineno = f.f_lineno
             co = f.f_code
             filename = co.co_filename
-            name = get_func_str_from_code_object(co)
-            file_descr = "".join([
-                '  ',
-                color("File ", color.fg_colors[0], bold=True), format_filename(filename), ", ",
-                color("line ", color.fg_colors[0]), color("%d" % lineno, color.fg_colors[4]), ", ",
-                color("in ", color.fg_colors[0]), name])
+            if not os.path.isfile(filename):
+                alt_fn = fallback_findfile(filename)
+                if alt_fn:
+                    filename = alt_fn
+            name = get_func_str_from_code_object(co, frame=f)
+            file_descr = "".join(
+                [
+                    "  ",
+                    color("File ", color.fg_colors[0], bold=True),
+                    format_filename(filename),
+                    ", ",
+                    color("line ", color.fg_colors[0]),
+                    color("%d" % lineno, color.fg_colors[4]),
+                    ", ",
+                    color("in ", color.fg_colors[0]),
+                    name,
+                ]
+            )
             with output.fold_text_ctx(file_descr):
-                if not os.path.isfile(filename):
-                    alt_fn = fallback_findfile(filename)
-                    if alt_fn:
-                        output(
-                            color("    -- couldn't find file, trying this instead: ", color.fg_colors[0]) +
-                            format_filename(alt_fn))
-                        filename = alt_fn
                 source_code = get_source_code(filename, lineno, f.f_globals)
                 if source_code:
                     source_code = remove_indent_lines(replace_tab_indents(source_code)).rstrip()
@@ -1164,7 +1193,7 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                     elif isinstance(f, DummyFrame) and not f.have_vars_available:
                         pass
                     else:
-                        with output.fold_text_ctx(color('    locals:', color.fg_colors[0])):
+                        with output.fold_text_ctx(color("    locals:", color.fg_colors[0])):
                             already_printed_locals = set()  # type: typing.Set[typing.Tuple[str,...]]
                             for token_str in grep_full_py_identifiers(parse_py_statement(source_code)):
                                 splitted_token = tuple(token_str.split("."))
@@ -1173,24 +1202,42 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
                                         continue
                                     token_value = None
                                     token_value = _try_set(
-                                        token_value, color("<local> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_locals, token)))
+                                        token_value,
+                                        color("<local> ", color.fg_colors[0]),
+                                        lambda: format_py_obj(_resolve_identifier(f.f_locals, token)),
+                                    )
                                     token_value = _try_set(
-                                        token_value, color("<global> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_globals, token)))
+                                        token_value,
+                                        color("<global> ", color.fg_colors[0]),
+                                        lambda: format_py_obj(_resolve_identifier(f.f_globals, token)),
+                                    )
                                     token_value = _try_set(
-                                        token_value, color("<builtin> ", color.fg_colors[0]),
-                                        lambda: format_py_obj(_resolve_identifier(f.f_builtins, token)))
+                                        token_value,
+                                        color("<builtin> ", color.fg_colors[0]),
+                                        lambda: format_py_obj(_resolve_identifier(f.f_builtins, token)),
+                                    )
                                     token_value = token_value or color("<not found>", color.fg_colors[0])
-                                    prefix = (
-                                        '      %s ' % color(".", color.fg_colors[0], bold=True).join(token) +
-                                        color("= ", color.fg_colors[0], bold=True))
+                                    prefix = "      %s " % color(".", color.fg_colors[0], bold=True).join(
+                                        token
+                                    ) + color("= ", color.fg_colors[0], bold=True)
                                     output(prefix, token_value)
                                     already_printed_locals.add(token)
                             if len(already_printed_locals) == 0:
                                 output(color("       no locals", color.fg_colors[0]))
                 else:
-                    output(color('    -- code not available --', color.fg_colors[0]))
+                    output(color("    -- code not available --", color.fg_colors[0]))
+
+            # Just like :func:`traceback.clear_frames`, but has an additional fix
+            # (https://github.com/python/cpython/issues/113939).
+            try:
+                f.clear()
+            except RuntimeError:
+                pass
+            else:
+                # Using this code triggers that the ref actually goes out of scope, otherwise it does not!
+                # https://github.com/python/cpython/issues/113939
+                f.f_locals  # noqa
+
             if isframe(_tb):
                 _tb = _tb.f_back
             elif is_stack_summary(_tb):
@@ -1204,6 +1251,7 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
     except Exception:
         output(color("ERROR: cannot get more detailed exception info because:", color.fg_colors[1], bold=True))
         import traceback
+
         for line in traceback.format_exc().split("\n"):
             output("   " + line)
 
@@ -1212,6 +1260,8 @@ def format_tb(tb=None, limit=None, allLocals=None, allGlobals=None, withTitle=Fa
 
 def print_tb(tb, file=None, **kwargs):
     """
+    Replacement for traceback.print_tb.
+
     :param types.TracebackType|types.FrameType|StackSummary tb:
     :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
     :return: nothing, prints to ``file``
@@ -1223,9 +1273,44 @@ def print_tb(tb, file=None, **kwargs):
     file.flush()
 
 
-def better_exchook(etype, value, tb,
-                   debugshell=False, autodebugshell=True,
-                   file=None, with_color=None, with_preamble=True):
+def print_exception(etype, value, tb, limit=None, file=None, chain=True):
+    """
+    Replacement for traceback.print_exception.
+
+    :param etype: exception type
+    :param value: exception value
+    :param tb: traceback
+    :param int|None limit:
+    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
+    :param bool chain: whether to print the chain of exceptions
+    """
+    better_exchook(etype, value, tb, autodebugshell=False, file=file, limit=limit, chain=chain)
+
+
+def print_exc(limit=None, file=None, chain=True):
+    """
+    Replacement for traceback.print_exc.
+    Shorthand for 'print_exception(*sys.exc_info(), limit, file)'.
+
+    :param int|None limit:
+    :param io.TextIOBase|io.StringIO|typing.TextIO|None file: stderr by default
+    :param bool chain:
+    """
+    print_exception(*sys.exc_info(), limit=limit, file=file, chain=chain)
+
+
+def better_exchook(
+    etype,
+    value,
+    tb,
+    debugshell=False,
+    autodebugshell=True,
+    file=None,
+    with_color=None,
+    with_preamble=True,
+    limit=None,
+    chain=True,
+):
     """
     Replacement for sys.excepthook.
 
@@ -1238,24 +1323,27 @@ def better_exchook(etype, value, tb,
         and exception information. stderr by default.
     :param bool|None with_color: whether to use ANSI escape codes for colored output
     :param bool with_preamble: print a short preamble for the exception
+    :param int|None limit:
+    :param bool chain: whether to print the chain of exceptions
     """
     if file is None:
         file = sys.stderr
 
     color = Color(enable=with_color)
-    output = _Output(color=color)
+    output = _OutputLinesCollector(color=color)
 
     rec_args = dict(autodebugshell=False, file=file, with_color=with_color, with_preamble=with_preamble)
-    if getattr(value, "__cause__", None):
-        better_exchook(type(value.__cause__), value.__cause__, value.__cause__.__traceback__, **rec_args)
-        output("")
-        output("The above exception was the direct cause of the following exception:")
-        output("")
-    elif getattr(value, "__context__", None):
-        better_exchook(type(value.__context__), value.__context__, value.__context__.__traceback__, **rec_args)
-        output("")
-        output("During handling of the above exception, another exception occurred:")
-        output("")
+    if chain:
+        if getattr(value, "__cause__", None):
+            better_exchook(type(value.__cause__), value.__cause__, value.__cause__.__traceback__, **rec_args)
+            output("")
+            output("The above exception was the direct cause of the following exception:")
+            output("")
+        elif getattr(value, "__context__", None):
+            better_exchook(type(value.__context__), value.__context__, value.__context__.__traceback__, **rec_args)
+            output("")
+            output("During handling of the above exception, another exception occurred:")
+            output("")
 
     def format_filename(s):
         """
@@ -1264,9 +1352,10 @@ def better_exchook(etype, value, tb,
         """
         base = os.path.basename(s)
         return (
-            color('"' + s[:-len(base)], color.fg_colors[2]) +
-            color(base, color.fg_colors[2], bold=True) +
-            color('"', color.fg_colors[2]))
+            color('"' + s[: -len(base)], color.fg_colors[2])
+            + color(base, color.fg_colors[2], bold=True)
+            + color('"', color.fg_colors[2])
+        )
 
     if with_preamble:
         output(color("EXCEPTION", color.fg_colors[1], bold=True))
@@ -1274,7 +1363,14 @@ def better_exchook(etype, value, tb,
     if tb is not None:
         output.lines.extend(
             format_tb(
-                tb=tb, allLocals=all_locals, allGlobals=all_globals, withTitle=True, with_color=color.enable))
+                tb=tb,
+                limit=limit,
+                allLocals=all_locals,
+                allGlobals=all_globals,
+                withTitle=True,
+                with_color=color.enable,
+            )
+        )
     else:
         output(color("better_exchook: traceback unknown", color.fg_colors[1]))
 
@@ -1283,17 +1379,24 @@ def better_exchook(etype, value, tb,
         # so do it in a similar way here as well.
         filename = value.filename
         # Keep the output somewhat consistent with format_tb.
-        file_descr = "".join([
-            '  ',
-            color("File ", color.fg_colors[0], bold=True), format_filename(filename), ", ",
-            color("line ", color.fg_colors[0]), color("%d" % value.lineno, color.fg_colors[4])])
+        file_descr = "".join(
+            [
+                "  ",
+                color("File ", color.fg_colors[0], bold=True),
+                format_filename(filename),
+                ", ",
+                color("line ", color.fg_colors[0]),
+                color("%d" % value.lineno, color.fg_colors[4]),
+            ]
+        )
         with output.fold_text_ctx(file_descr):
             if not os.path.isfile(filename):
                 alt_fn = fallback_findfile(filename)
                 if alt_fn:
                     output(
-                        color("    -- couldn't find file, trying this instead: ", color.fg_colors[0]) +
-                        format_filename(alt_fn))
+                        color("    -- couldn't find file, trying this instead: ", color.fg_colors[0])
+                        + format_filename(alt_fn)
+                    )
                     filename = alt_fn
             source_code = get_source_code(filename, value.lineno)
             if source_code:
@@ -1304,7 +1407,7 @@ def better_exchook(etype, value, tb,
                 indent_prefix = get_same_indent_prefix(lines)
                 if indent_prefix is None:
                     indent_prefix = ""
-                source_code = "".join([line[len(indent_prefix):] for line in lines])
+                source_code = "".join([line[len(indent_prefix) :] for line in lines])
                 source_code = source_code.rstrip()
                 prefix = "    line: "
                 output(prefix, color.py_syntax_highlight(source_code), color=color.fg_colors[0])
@@ -1322,7 +1425,7 @@ def better_exchook(etype, value, tb,
         try:
             return str(value)
         except Exception:
-            return '<unprintable %s object>' % type(value).__name__
+            return "<unprintable %s object>" % type(value).__name__
 
     # noinspection PyShadowingNames
     def _format_final_exc_line(etype, value):
@@ -1334,9 +1437,12 @@ def better_exchook(etype, value, tb,
         return line
 
     # noinspection PyUnresolvedReferences
-    if (isinstance(etype, BaseException) or
-            (hasattr(types, "InstanceType") and isinstance(etype, types.InstanceType)) or
-            etype is None or type(etype) is str):
+    if (
+        isinstance(etype, BaseException)
+        or (hasattr(types, "InstanceType") and isinstance(etype, types.InstanceType))
+        or etype is None
+        or type(etype) is str
+    ):
         output(_format_final_exc_line(etype, value))
     else:
         output(_format_final_exc_line(etype.__name__, value))
@@ -1385,7 +1491,7 @@ def dump_all_thread_tracebacks(exclude_thread_ids=None, file=None):
             thread = threads.get(tid)
             if thread:
                 assert isinstance(thread, threading.Thread)
-                if thread is threading.currentThread():
+                if thread is threading.current_thread():
                     tags += ["current"]
                 # noinspection PyProtectedMember,PyUnresolvedReferences
                 if isinstance(thread, threading._MainThread):
@@ -1417,13 +1523,14 @@ def get_current_frame():
         return sys.exc_info()[2].tb_frame.f_back
 
 
-def get_func_str_from_code_object(co):
+def get_func_str_from_code_object(co, frame=None):
     """
     :param types.CodeType co:
+    :param types.FrameType|DummyFrame|None frame: if given, might provide a faster way to get the function name
     :return: co.co_name as fallback, but maybe sth better like the full func name if possible
     :rtype: str
     """
-    f = get_func_from_code_object(co)
+    f = get_func_from_code_object(co, frame=frame)
     if f:
         if hasattr(f, "__qualname__"):
             return f.__qualname__
@@ -1431,26 +1538,72 @@ def get_func_str_from_code_object(co):
     return co.co_name
 
 
-def get_func_from_code_object(co):
+def get_func_from_code_object(co, frame=None):
     """
     :param types.CodeType co:
+    :param types.FrameType|DummyFrame|None frame: if given, might provide a faster way to get the function name
     :return: function, such that ``func.__code__ is co``, or None
     :rtype: types.FunctionType
 
     This is CPython specific (to some degree; it uses the `gc` module to find references).
-    Inspired from: https://stackoverflow.com/questions/12787108/getting-the-python-function-for-a-code-object
+    Inspired from:
+    https://stackoverflow.com/questions/12787108/getting-the-python-function-for-a-code-object
+    https://stackoverflow.com/questions/54656758/get-function-object-from-stack-frame-object
     """
-    import types
-    if not isinstance(co, types.CodeType):
-        return None
     import gc
+    import types
+
+    assert isinstance(co, (types.CodeType, DummyFrame))
+    _attr_name = "__code__" if PY3 else "func_code"
+    if frame and frame.f_code.co_nlocals > 0:
+        func_name = frame.f_code.co_name
+        frame_self = frame.f_locals.get("self")
+        if frame_self is not None:
+            candidate = getattr(frame_self.__class__, func_name, None)
+            if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
+                return candidate
+    try:
+        candidate = getattr(_get_loaded_module_from_filename(co.co_filename), co.co_name, None)
+    except ImportError:  # some modules have lazy loaders, but those might fail here
+        candidate = None
+    if candidate and (getattr(candidate, _attr_name, None) is co or isinstance(co, DummyFrame)):
+        return candidate
+    if isinstance(co, DummyFrame):
+        return None
     candidates = gc.get_referrers(co)
-    candidates = [f for f in candidates if getattr(f, "__code__" if PY3 else "func_code", None) is co]
+    candidates = [f for f in candidates if getattr(f, _attr_name, None) is co]
     if candidates:
         return candidates[0]
     return None
 
 
+_loaded_module_from_filename_cache = {}  # filename -> module name
+
+
+def _get_loaded_module_from_filename(filename):
+    """
+    Like inspect.getmodule but faster.
+
+    :param str filename:
+    :rtype: types.ModuleType|Any|None
+    """
+    if filename.endswith(".pyc") or filename.endswith(".pyo"):
+        filename = filename[:-1]
+    if filename in _loaded_module_from_filename_cache:
+        return sys.modules.get(_loaded_module_from_filename_cache[filename])
+    # Update the filename to module name cache and check yet again
+    # Copy sys.modules in order to cope with changes while iterating
+    for modname, module in sys.modules.copy().items():
+        f = getattr(module, "__file__", None)
+        if f:
+            if f.endswith(".pyc") or f.endswith(".pyo"):
+                f = f[:-1]
+            _loaded_module_from_filename_cache[f] = modname
+    if filename in _loaded_module_from_filename_cache:
+        return sys.modules.get(_loaded_module_from_filename_cache[filename])
+    return None
+
+
 def iter_traceback(tb=None, enforce_most_recent_call_first=False):
     """
     Iterates a traceback of various formats:
@@ -1513,6 +1666,7 @@ class ExtendedFrameSummary(FrameSummary):
     """
     Extends :class:`FrameSummary` by ``self.tb_frame``.
     """
+
     def __init__(self, frame, **kwargs):
         super(ExtendedFrameSummary, self).__init__(**kwargs)
         self.tb_frame = frame
@@ -1544,7 +1698,12 @@ class DummyFrame:
         self.f_locals = f_locals or {}
         self.f_globals = f_globals or {}
         self.f_builtins = f_builtins or {}
-        self.have_vars_available = (f_locals is not None or f_globals is not None or f_builtins is not None)
+        self.have_vars_available = f_locals is not None or f_globals is not None or f_builtins is not None
+        self.co_nlocals = len(self.f_locals) if self.f_locals is not self.f_globals else 0
+
+    def clear(self):
+        """clear"""
+        self.f_locals = None
 
 
 # noinspection PyPep8Naming,PyUnusedLocal
@@ -1572,8 +1731,7 @@ def _StackSummary_extract(frame_gen, limit=None, lookup_lines=True, capture_loca
         co = f.f_code
         filename = co.co_filename
         name = co.co_name
-        result.append(ExtendedFrameSummary(
-            frame=f, filename=filename, lineno=lineno, name=name, lookup_line=False))
+        result.append(ExtendedFrameSummary(frame=f, filename=filename, lineno=lineno, name=name, lookup_line=False))
     return result
 
 
@@ -1596,7 +1754,39 @@ def replace_traceback_format_tb():
     and is not officially supported by Python.
     """
     import traceback
+
     traceback.format_tb = format_tb
     if hasattr(traceback, "StackSummary"):
         traceback.StackSummary.format = format_tb
         traceback.StackSummary.extract = _StackSummary_extract
+
+
+def replace_traceback_print_tb():
+    """
+    Replaces these functions from the traceback module by our own:
+
+    - traceback.print_tb
+    - traceback.print_exception
+    - traceback.print_exc
+
+    Note that this kind of monkey patching might not be safe under all circumstances
+    and is not officially supported by Python.
+    """
+    import traceback
+
+    traceback.print_tb = print_tb
+    traceback.print_exception = print_exception
+    traceback.print_exc = print_exc
+
+
+def setup_all():
+    """
+    Calls:
+
+    - :func:`install`
+    - :func:`replace_traceback_format_tb`
+    - :func:`replace_traceback_print_tb`
+    """
+    install()
+    replace_traceback_format_tb()
+    replace_traceback_print_tb()
